/*
 * Overseer
 * Planets player automation utility
 * by Roger Burton West <rburtonw@nyx.net>
 * GNUWare
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <math.h>
#include <unistd.h>
#include <tkplayer.h>

enum { std, in_macro, in_match, transition, in_action };

enum { done, qcomp, scomp, multiq };

bool shipmatch[501];
bool matchedthis[501];
bool gravwell=true;
int playerid;

SHIPDATA *shipdata;
PLANETDATA *planetdata;
BASEDATA *basedata;
PLANET_COORDS *planet_coords;

XString macroname[50];
XString macrocont[50];
int macros=0;

void help()
{
  fprintf(stderr,"Usage: overseer <dir> <playernum\n");
  fprintf(stderr,"See documentation for more info!\n");
}

int find_planet(int x,int y)
{
  int plan=0;
  for (int v=1;v<=500;v++) {
    if ((planet_coords[v].x==x) &&
        (planet_coords[v].y==y)) {
      plan=v;
      v=501;
    }
  }
  return plan;
}

bool lcompare(long value,XString tests) {
  bool k;
  int v;
  char ch,ch2;
  long testval;

  if (tests == "") return false;

  v=tests.Find(",");
  if (v != -1) {
    k=lcompare(value,tests.Left(v));
    if (!k) k=lcompare(value,tests.Mid(v+1));
    return k;
  }

  ch=tests.GetAt(0);
  ch2=tests.GetAt(1);
  switch (ch) {
    case '<':
    case '>':
      switch(ch2) {
        case '=':
          tests=tests.Mid(2);
          break;
        default:
          tests=tests.Mid(1);
      }
      break;
    case '=':
      tests=tests.Mid(1);
      break;
    default:
      ch='=';
  }

  testval=atol((const char *)tests);

  switch (ch) {
    case '<':
      if (ch2=='=') {
        k=(value <= testval);
      } else {
        k=(value < testval);
      }
      break;
    case '>':
      if (ch2=='=') {
        k=(value >= testval);
      } else {
        k=(value > testval);
      }
      break;
    default:
      k=(value == testval);
  }

  return k;
}

bool scompare(XString value,XString tests) {
  bool k;
  int v;

  if (tests == "") return false;

  v=tests.Find(",");
  if (v != -1) {
    k=scompare(value,tests.Left(v));
    if (!k) k=scompare(value,tests.Mid(v+1));
    return k;
  }

  value=value.Left(tests.Len());

  k=(value == tests);

  return k;
}

void process_line(XString line)
{
  static int state=std;
  static int matchcount=0;
  static int mc=0;
  static int shipact[501];
  XString word[20];
  int wordc,words;

  if (line == "") return;

  wordc=line.Find("'");
  if (wordc != -1) {
    line=line.Left(wordc);
  }

  wordc=line.Find(";");
  if (wordc >= 0) {
    while ((wordc=line.Find(";")) != -1) {
      process_line(line.Left(wordc));
      line=line.Mid(wordc+1);
      line.Trim();
    };
    process_line(line);
    return;
  }

  for (int p=0;p<5;p++) {
    char k[3];
    k[1]=' ';
    k[2]=0;
    switch (p) {
      case 0:
        k[0]='<';
        break;
      case 1:
        k[0]='>';
        break;
      case 2:
        k[0]='=';
        break;
      case 3:
        k[0]=',';
        break;
      default:
        k[0]=' ';
    }
    while ((wordc=line.Find(k)) != -1) {
      line=line.Left(wordc+1) | line.Mid(wordc+2);
      line.Trim();
    };
  }

  line.Trim();

  words=0;
  while ((wordc=line.Find(" ")) != -1) {
    word[words++]=line.Left(wordc);
    line=line.Mid(wordc+1);
    line.Trim();
  };
  word[words++]=line;

  if (macros>0) {
    for (int k=0;k<macros;k++) {
      if (word[0]==macroname[k]) {
        process_line(macrocont[k]);
        return;
      }
    }
  }

  if (-word[0]=="config") {
    if (-word[1]=="gravwell") {
      if (-word[2]=="on") {
        gravwell=true;
      } else {
        gravwell=false;
      }
    }
  }

  switch (state) {
    case std:
      if (-word[0]=="macdef" && word[2]=="{") {
        state=in_macro;
        macroname[macros]=word[1];
      }
      if (-word[0]=="match" && word[1]=="{") {
        state=in_match;
        matchcount=0;
        for (int k=1;k<=500;k++) {
          matchedthis[k]=!shipmatch[k];
          if (matchedthis[k]) matchcount++;
        }
      }
      break;
    case in_macro:
      if (word[0]=="}") {
        macros++;
        state=std;
      } else {
        macrocont[macros]=macrocont[macros] | line | ";";
      }
      break;
    case in_match:
      if (word[0]=="}") {
        state=std;
        if (matchcount>0) {
          for (int k=1;k<=500;k++) {
            if (matchedthis[k]) shipmatch[k]=true;
          }
          state=transition;
        }
      } else {
        if (matchcount>0) {
          for (int k=1;k<=500;k++) {
            if (matchedthis[k]) {
              long qu1=-1,qu2=-1;
              XString su1;
              int cmode=done;
              if (-word[0]=="atship") {
                int os;
                os=atoi((const char *)word[1]);
                if (shipdata[os].wID != 0) {
                  word[0]="at";
                  word[1]=shipdata[os].xCoord;
                  word[2]=shipdata[os].yCoord;
                  words=3;
                }
              }
              if (-word[0]=="at") {
                switch (words) {
                  case 2:
                    int plan;
                    plan=find_planet(shipdata[k].xCoord,shipdata[k].yCoord);
                    if (plan>0) {
                      cmode=qcomp;
                      qu1=plan;
                    } else {
                      matchedthis[k]=false;
                    }
                    break;
                  case 3:
                    cmode=multiq;
                    qu1=shipdata[k].xCoord;
                    qu2=shipdata[k].yCoord;
                    break;
                  case 5:
                    if (shipdata[k].xCoord < atoi((const char *)word[1])) matchedthis[k]=false;
                    if (shipdata[k].yCoord < atoi((const char *)word[2])) matchedthis[k]=false;
                    if (shipdata[k].xCoord > atoi((const char *)word[3])) matchedthis[k]=false;
                    if (shipdata[k].yCoord > atoi((const char *)word[4])) matchedthis[k]=false;
                    break;
                }
              }
              if (-word[0]=="class") {
                cmode=qcomp;
                qu1=shipdata[k].wShipType;
              }
              if (-word[0]=="id") {
                cmode=qcomp;
                qu1=k;
              }
              if (-word[0]=="range") {
                cmode=qcomp;
                int d;
                d=shipdata[k].xDis*shipdata[k].xDis+shipdata[k].yDis*shipdata[k].yDis;
                d=(int)sqrt(d);
                qu1=d;
              }
              if (-word[0]=="eta") {
                cmode=qcomp;
                int s,d;
                d=shipdata[k].xDis*shipdata[k].xDis+shipdata[k].yDis*shipdata[k].yDis;
                d=(int)sqrt(d);
                if (d==0) {
                  qu1=0;
                } else {
                  s=shipdata[k].wWarp;
                  s=s*s;
                  if (s==0) {
                    qu1=999;
                  } else {
                    qu1=d/s+1;
                  }
                }
              }
              if (-word[0]=="nengine") {
                cmode=qcomp;
                HULLSPEC *hullspec;
                hullspec=read_hullspec(shipdata[k].wShipType);
                qu1=hullspec->wEngines;
                delete hullspec;
              }
              if (-word[0]=="engine") {
                cmode=qcomp;
                qu1=shipdata[k].wDriveLevel;
              }
              if (-word[0]=="nbeam") {
                cmode=qcomp;
                qu1=shipdata[k].wBeamWeapons;
              }
              if (-word[0]=="beam") {
                cmode=qcomp;
                qu1=shipdata[k].wBeamLevel;
              }
              if (-word[0]=="ntorp") {
                cmode=qcomp;
                qu1=shipdata[k].wTorpsBays;
              }
              if (-word[0]=="torp") {
                cmode=qcomp;
                qu1=shipdata[k].wTorpLevel;
              }
              if (-word[0]=="nbay") {
                cmode=qcomp;
                qu1=shipdata[k].wFighterBays;
              }
              if (-word[0]=="mission") {
                cmode=qcomp;
                qu1=shipdata[k].wMission;
              }
              if (-word[0]=="damage") {
                cmode=qcomp;
                qu1=shipdata[k].wDamage;
              }
              if (-word[0].Left(1)=="p") {
                int plan;
                plan=find_planet(shipdata[k].xCoord,shipdata[k].yCoord);
                if (plan>0) {
                  if (planetdata[plan].wID != plan) {
                    plan=0;
                  }
                }
                if (plan==0) {
                  matchedthis[k]=false;
                } else {
                  cmode=qcomp;
                  if (-word[0]=="ptemp") {
                    qu1=100-planetdata[plan].wPlanet;
                  }
                  if (-word[0]=="pcpop") {
                    qu1=planetdata[plan].lColonists;
                  }
                  if (-word[0]=="powner") {
                    qu1=planetdata[plan].wOwner;
                  }
                  if (-word[0]=="pnpop") {
                    qu1=planetdata[plan].lNatives;
                  }
                  if (-word[0]=="pntype") {
                    qu1=planetdata[plan].wRace;
                  }
                }
              }
              if (-word[0]=="cfree") {
                cmode=qcomp;
                HULLSPEC *hullspec;
                hullspec=read_hullspec(shipdata[k].wShipType);
                qu1=hullspec->wCargo;
                delete hullspec;
                qu1 -= shipdata[k].wTri;
                qu1 -= shipdata[k].wDur;
                qu1 -= shipdata[k].wMol;
                qu1 -= shipdata[k].wSupplies;
                qu1 -= shipdata[k].wColonists;
                qu1 -= shipdata[k].wTorps;
              }
              if (-word[0]=="ctotal") {
                cmode=qcomp;
                HULLSPEC *hullspec;
                hullspec=read_hullspec(shipdata[k].wShipType);
                qu1=hullspec->wCargo;
                delete hullspec;
              }
              if (-word[0]=="fcode") {
                cmode=scomp;
                su1=shipdata[k].sCode;
              }
              if (-word[0]=="name") {
                cmode=scomp;
                su1=shipdata[k].sName;
              }
              if (-word[0]=="cargo") {
                cmode=qcomp;
                qu1=-1;
                if (-word[1].Left(1)=="n") {
                  qu1=shipdata[k].wFuel;
                }
                if (-word[1].Left(1)=="t") {
                  qu1=shipdata[k].wTri;
                }
                if (-word[1].Left(1)=="d") {
                  qu1=shipdata[k].wDur;
                }
                if (-word[1].Left(1)=="m") {
                  qu1=shipdata[k].wMol;
                }
                if (-word[1].Left(1)=="s") {
                  qu1=shipdata[k].wSupplies;
                }
                if (-word[1].Left(1)=="c") {
                  qu1=shipdata[k].wColonists;
                }
                if (-word[1].Left(1)=="$" || -word[1]=="money" || -word[1]=="mc") {
                  qu1=shipdata[k].wMoney;
                }
                if (-word[1].Left(1)=="o" || -word[1]=="torp") {
                  if (shipdata[k].wTorpsBays>0) {
                    qu1=shipdata[k].wTorps;
                  } else {
                    qu1=0;
                  }
                }
                if (-word[1].Left(1)=="f") {
                  if (shipdata[k].wFighterBays>0) {
                    qu1=shipdata[k].wTorps;
                  } else {
                    qu1=0;
                  }
                }
                if (qu1==-1) {
                  cmode=done;
                } else {
                  word[1]=word[2];
                }
              }
              switch (cmode) {
                case multiq:
                  matchedthis[k]=lcompare(qu2,word[2]);
                case qcomp:
                  if (matchedthis[k]) matchedthis[k]=lcompare(qu1,word[1]);
                  break;
                case scomp:
                  matchedthis[k]=scompare(su1,word[1]);
                  break;
              }
              if (!matchedthis[k]) {
                matchcount--;
              }
            }
          }
        }
      }
      break;
    case transition:
      if (-word[0]=="action" && word[1]=="{") {
        state=in_action;
        mc=0;
        for (int k=1;k<=500;k++) {
          if (matchedthis[k]) {
            shipact[mc++]=k;
          }
        }
      }
      break;
    case in_action:
      if (word[0]=="}") {
        state=std;
      } else {
        for (int mt=0;mt<mc;mt++) {
          int k=shipact[mt];
          bool ok=false;
          if (-word[0]=="goto") {
            int gx=0,gy=0;
            bool set=false;
            switch (words) {
              case 2:
                if (-word[1]=="base") {
                  long dmax=999999;
                  long d,d2;
                  for (int v=1;v<501;v++) {
                    if (basedata[v].wID > 0) {
                      d  =(planet_coords[v].x-shipdata[k].xCoord);
                      d *= d;
                      d2 =(planet_coords[v].y-shipdata[k].yCoord);
                      d2 *= d2;
                      d += d2;
                      if (d<dmax) {
                        dmax=d;
                        gx=planet_coords[v].x;
                        gy=planet_coords[v].y;
                      }
                    }
                  }
                } else {
                  if (-word[1]=="planet") {
                    long dmax=999999;
                    long d,d2;
                    for (int v=1;v<501;v++) {
                      if (planetdata[v].wOwner == playerid) {
                        d  =(planet_coords[v].x-shipdata[k].xCoord);
                        d *= d;
                        d2 =(planet_coords[v].y-shipdata[k].yCoord);
                        d2 *= d2;
                        d += d2;
                        if (d<dmax) {
                          dmax=d;
                          gx=planet_coords[v].x;
                          gy=planet_coords[v].y;
                        }
                      }
                    }
                  } else {
                    int v;
                    v=atoi((const char *)word[1]);
                    gx=planet_coords[v].x;
                    gy=planet_coords[v].y;
                  }
                }
                gx-=shipdata[k].xCoord;
                gy-=shipdata[k].yCoord;
                if (gravwell) {
                  if (gx> 2) gx-=2;
                  if (gx<-2) gx+=2;
                  if (gy> 2) gy-=2;
                  if (gy<-2) gy+=2;
                }
                set=true;
                break;
              case 3:
                gx=atoi((const char *)word[1])-shipdata[k].xCoord;
                gy=atoi((const char *)word[2])-shipdata[k].yCoord;
                set=true;
                break;
            }
            if (set) {
              shipdata[k].xDis=gx;
              shipdata[k].yDis=gy;
              ok=true;
            }
          }
          if (-word[0]=="load") {
            int qty=0;
            long qttemp=0;
            int plan;
            plan=find_planet(shipdata[k].xCoord,shipdata[k].yCoord);
            if (plan!=0 && planetdata[plan].wOwner == playerid) {
              if (-word[1]=="all") {
                qty=99999;
              } else {
                qty=atoi((const char *)word[1]);
              }
              HULLSPEC *hullspec;
              hullspec=read_hullspec(shipdata[k].wShipType);
              if (-word[2].Left(1)=="n") {
                qttemp=hullspec->wFuel;
              } else {
                qttemp=hullspec->wCargo;
                qttemp -= shipdata[k].wTri;
                qttemp -= shipdata[k].wDur;
                qttemp -= shipdata[k].wMol;
                qttemp -= shipdata[k].wSupplies;
                qttemp -= shipdata[k].wColonists;
                qttemp -= shipdata[k].wTorps;
              }
              delete hullspec;
              if (qttemp < qty) qty=qttemp;
              if (-word[2].Left(1)=="n") {
                qttemp=planetdata[plan].lFuel;
                if (qttemp < qty) qty=qttemp;
                planetdata[plan].lFuel -= qty;
                shipdata[k].wFuel += qty;
                write_planetdata(plan,&planetdata[plan]);
              }
              if (-word[2].Left(1)=="t") {
                qttemp=planetdata[plan].lTri;
                if (qttemp < qty) qty=qttemp;
                planetdata[plan].lTri -= qty;
                shipdata[k].wTri += qty;
                write_planetdata(plan,&planetdata[plan]);
              }
              if (-word[2].Left(1)=="d") {
                qttemp=planetdata[plan].lDur;
                if (qttemp < qty) qty=qttemp;
                planetdata[plan].lDur -= qty;
                shipdata[k].wDur += qty;
                write_planetdata(plan,&planetdata[plan]);
              }
              if (-word[2].Left(1)=="m") {
                qttemp=planetdata[plan].lMol;
                if (qttemp < qty) qty=qttemp;
                planetdata[plan].lMol -= qty;
                shipdata[k].wMol += qty;
                write_planetdata(plan,&planetdata[plan]);
              }
              if (-word[2].Left(1)=="s") {
                qttemp=planetdata[plan].lSupplies;
                if (qttemp < qty) qty=qttemp;
                planetdata[plan].lSupplies -= qty;
                shipdata[k].wSupplies += qty;
                write_planetdata(plan,&planetdata[plan]);
              }
              if (-word[2].Left(1)=="c") {
                qttemp=planetdata[plan].lColonists;
                if (qttemp < qty) qty=qttemp;
                planetdata[plan].lColonists -= qty;
                shipdata[k].wColonists += qty;
                write_planetdata(plan,&planetdata[plan]);
              }
              if (-word[2].Left(1)=="$" || -word[2]=="money" || -word[2]=="mc") {
                qttemp=planetdata[plan].lMoney;
                if (qttemp < qty) qty=qttemp;
                planetdata[plan].lMoney -= qty;
                shipdata[k].wMoney += qty;
                write_planetdata(plan,&planetdata[plan]);
              }
              if (-word[2].Left(1)=="o" || -word[2]=="torp") {
                if (basedata[plan].wID == plan) {
                  if (shipdata[k].wTorpsBays>0) {
                    qttemp=basedata[plan].wTorps[shipdata[k].wTorpLevel];
                    if (qttemp < qty) qty=qttemp;
                    basedata[plan].wTorps[shipdata[k].wTorpLevel-1] -= qty;
                    shipdata[k].wTorps += qty;
                    write_basedata(plan,&basedata[plan]);
                  }
                }
              }
              if (-word[2].Left(1)=="f") {
                if (basedata[plan].wID == plan) {
                  if (shipdata[k].wFighterBays>0) {
                    qttemp=basedata[plan].wFighters;
                    if (qttemp < qty) qty=qttemp;
                    basedata[plan].wFighters -= qty;
                    shipdata[k].wTorps += qty;
                    write_basedata(plan,&basedata[plan]);
                  }
                }
              }
              ok=true;
            }
          }
          if (-word[0]=="unload") {
            int qty=0;
            long qttemp=0;
            int plan;
            plan=find_planet(shipdata[k].xCoord,shipdata[k].yCoord);
            if (plan!=0) {
              if (-word[1]=="all") {
                qty=99999;
              } else {
                qty=atoi((const char *)word[1]);
              }
              if (-word[2].Left(1)=="n") {
                qttemp=shipdata[k].wFuel;
                if (qttemp < qty) qty=qttemp;
                planetdata[plan].lFuel += qty;
                shipdata[k].wFuel -= qty;
                write_planetdata(plan,&planetdata[plan]);
              }
              if (-word[2].Left(1)=="t") {
                qttemp=shipdata[k].wTri;
                if (qttemp < qty) qty=qttemp;
                planetdata[plan].lTri += qty;
                shipdata[k].wTri -= qty;
                write_planetdata(plan,&planetdata[plan]);
              }
              if (-word[2].Left(1)=="d") {
                qttemp=shipdata[k].wDur;
                if (qttemp < qty) qty=qttemp;
                planetdata[plan].lDur += qty;
                shipdata[k].wDur -= qty;
                write_planetdata(plan,&planetdata[plan]);
              }
              if (-word[2].Left(1)=="m") {
                qttemp=shipdata[k].wMol;
                if (qttemp < qty) qty=qttemp;
                planetdata[plan].lMol += qty;
                shipdata[k].wMol -= qty;
                write_planetdata(plan,&planetdata[plan]);
              }
              if (-word[2].Left(1)=="s") {
                qttemp=shipdata[k].wSupplies;
                if (qttemp < qty) qty=qttemp;
                planetdata[plan].lSupplies += qty;
                shipdata[k].wSupplies -= qty;
                write_planetdata(plan,&planetdata[plan]);
              }
              if (-word[2].Left(1)=="c") {
                qttemp=shipdata[k].wColonists;
                if (qttemp < qty) qty=qttemp;
                if (planetdata[plan].lColonists == 0) {
                  shipdata[k].wPClans=qty;
                } else {
                  planetdata[plan].lColonists += qty;
                  write_planetdata(plan,&planetdata[plan]);
                }
                shipdata[k].wColonists -= qty;
              }
              if (-word[2].Left(1)=="$" || -word[2]=="money" || -word[2]=="mc") {
                qttemp=shipdata[k].wMoney;
                if (qttemp < qty) qty=qttemp;
                planetdata[plan].lMoney += qty;
                shipdata[k].wMoney -= qty;
                write_planetdata(plan,&planetdata[plan]);
              }
              if (-word[2].Left(1)=="o" || -word[2]=="torp") {
                if (basedata[plan].wID == plan) {
                  if (shipdata[k].wTorpsBays>0) {
                    qttemp=shipdata[k].wTorps;
                    if (qttemp < qty) qty=qttemp;
                    basedata[plan].wTorps[shipdata[k].wTorpLevel-1] += qty;
                    shipdata[k].wTorps -= qty;
                    write_basedata(plan,&basedata[plan]);
                  }
                }
              }
              if (-word[2].Left(1)=="f") {
                if (basedata[plan].wID == plan) {
                  if (shipdata[k].wFighterBays>0) {
                    qttemp=shipdata[k].wTorps;
                    if (qttemp < qty) qty=qttemp;
                    qttemp=60-basedata[plan].wFighters;
                    if (qttemp < qty) qty=qttemp;
                    basedata[plan].wFighters += qty;
                    shipdata[k].wTorps -= qty;
                    write_basedata(plan,&basedata[plan]);
                  }
                }
              }
              ok=true;
            }
          }
          if (-word[0]=="speed") {
            ok=true;
            int speed=0;
            if (-word[1]=="max") {
              speed=shipdata[k].wDriveLevel;
            } else {
              speed=atoi((const char *)word[1]);
            }
            shipdata[k].wWarp=speed;
          }
          if (-word[0]=="mission") {
            ok=true;
            int miss=1;
            miss=atoi((const char *)word[1]);
            shipdata[k].wMission=miss;
            if (miss==7) { //Tow
              shipdata[k].wTowed=atoi((const char *)word[2]);
            }
            if (miss==8) { //Intercept
              shipdata[k].wIntercept=atoi((const char *)word[2]);
            }
          }
          if (-word[0]=="tow") {
            ok=true;
            shipdata[k].wMission=7;
            shipdata[k].wTowed=atoi((const char *)word[1]);
          }
          if (-word[0]=="intercept") {
            ok=true;
            shipdata[k].wMission=8;
            shipdata[k].wIntercept=atoi((const char *)word[1]);
          }
          if (-word[0]=="enemy") {
            ok=true;
            shipdata[k].wEnemy=atoi((const char *)word[1]);
          }
          if (-word[0]=="fcode") {
            shipdata[k].sCode=word[1].Left(3);
            ok=true;
          }
          if (-word[0]=="name") {
            shipdata[k].sName=word[1].Left(20);
            ok=true;
          }
          if (-word[0]=="unmatch") {
            ok=true;
            shipmatch[k]=false;
          }
          if (ok) {
            write_shipdata(k,&shipdata[k]);
          }
        }
      }
  }
}

int main(int argc,char *argv[])
{
  bool parm_ok=false;
  FILE *handle;

  srandom(time(NULL));

  fprintf(stderr,"OVERSEER v0.2 by Roger Burton West <roger@firedrake.demon.co.uk>\n");

  if (argc == 3) {
    parm_ok=set_game_dir(argv[1]);
    playerid=atoi(argv[2]);
    if (parm_ok) parm_ok=set_player(playerid);
  }

  if (!parm_ok) {
    help();
    return 1;
  }

  handle=mkopen("over&p.ini",false,'r');

  if (handle) {

    shipdata=read_all_shipdata();
    planetdata=read_all_planetdata();
    basedata=read_all_basedata();
    planet_coords=read_all_planet_coords();

    for (int k=1;k<=500;k++) {
      shipmatch[k]=(shipdata[k].wID == 0);
    }

    do {
      int ch;
      char line[200];
      ch=-1;
      do {
        ch++;
        fread(&line[ch],1,1,handle);
      } while (line[ch]>=' ');
      do {
        fread(&line[ch],1,1,handle);
      } while (line[ch]<' ' && !feof(handle));
      if (!feof(handle)) fseek(handle,-1,SEEK_CUR);
      line[ch]=0;
      process_line((XString)line);
    } while (!feof(handle));
    fclose(handle);

    delete [](planet_coords);

    make_checksum(true,true,true);

  }

  return 0;
}
